#include <stdint.h>
#include <string.h>
#include "app_error.h"
#include "app_timer.h"
#include "app_util_platform.h"
#include "nordic_common.h"
#include "nrf_gpio.h"
#include "ble.h"
#include "ble_advdata.h"
#include "ble_advertising.h"
#include "ble_bas.h"
#include "ble_conn_params.h"
#include "ble_dis.h"
#include "ble_hci.h"
#include "ble_hts.h"
#include "ble_srv_common.h"
#include "ble_mds.h"
#include "nrf.h"
#include "rtos.h"
#include "softdevice_handler.h"
#include "stack.h"
#include "uc_config.h"

#define IS_SRVC_CHANGED_CHARACT_PRESENT  0                                          /**< Include or not the service_changed characteristic. if not enabled, the server's database cannot be changed for the lifetime of the device*/
#define TEMP_TYPE_AS_CHARACTERISTIC      0                                          /**< Determines if temperature type is given as characteristic (1) or as a field of measurement (0). */

#define DEVICE_NAME        "Who Knows"                                              /**< Name of device. Will be included in the advertising data. */
#define MANUFACTURER_NAME  "Someone"                                                /**< Manufacturer. Will be passed to Device Information Service. */
#define MODEL_NUM          "????0"                                                  /**< Model number. Will be passed to Device Information Service. */
#define MANUFACTURER_ID    0x0123456789ABCDEF                                       /**< Manufacturer ID, part of System ID. Will be passed to Device Information Service. */
#define ORG_UNIQUE_ID      0x26021964                                               /**< Organizational Unique ID, part of System ID. Will be passed to Device Information Service. */

#define MAX_TIMERS                       2                                      // max. number of simultaneously created timers
#define TIMER_PRESC                      0                                      // value of the RTC1 PRESCALER register
#define TIMER_QUEUE_SIZE                 4                                      // size of timer operation queues

#define APP_ADV_INTERVAL                512                                         /**< The advertising interval (in units of 0.625 ms. This value corresponds to 40 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      180                                         /**< The advertising timeout (in units of seconds). */

#define APP_TIMER_PRESCALER             0                                           /**< Value of the RTC1 PRESCALER register. */

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(500, UNIT_1_25_MS)            /**< Minimum acceptable connection interval (0.5 seconds). */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(1000, UNIT_1_25_MS)           /**< Maximum acceptable connection interval (1 second). */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)             /**< Connection supervisory timeout (4 seconds). */
#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(5000, APP_TIMER_PRESCALER)  /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(30000, APP_TIMER_PRESCALER) /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                           /**< Number of attempts before giving up the connection parameter negotiation. */

#define SEC_PARAM_BOND                  1                                           /**< Perform bonding. */
#define SEC_PARAM_MITM                  0                                           /**< Man In The Middle protection not required. */
#define SEC_PARAM_IO_CAPABILITIES       BLE_GAP_IO_CAPS_NONE                        /**< No I/O capabilities. */
#define SEC_PARAM_OOB                   0                                           /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE          7                                           /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE          16                                          /**< Maximum encryption key size. */

#define DEAD_BEEF                       0xDEADBEEF                                  /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

static ble_gap_sec_params_t             m_sec_params;                               /**< Security requirements for this application. */
static uint16_t                         m_conn_handle = BLE_CONN_HANDLE_INVALID;    /**< Handle of the current connection. */
static ble_bas_t                        m_bas;                                      /**< Structure used to identify the battery service. */
static ble_hts_t                        m_hts;                                      /**< Structure used to identify the health thermometer service. */
static ble_mds_t                        m_mds;

static bool                             m_hts_meas_ind_conf_pending = false;        /**< Flag to keep track of when an indication confirmation is pending. */

static uint32_t        celsiusX100 = 2602;

ble_uuid_t m_adv_uuids[] = {
                               { BLE_MDS_UUID_SERVICE, BLE_UUID_TYPE_VENDOR_BEGIN }
                           };                                                       /**< Universally unique service identifiers. */

static OS_TASK mTaskCb;

static OS_STACKPTR int32_t Stack[256];

static void ble_stack_init   (void);              /* Initialisation routines */
static void gap_params_init  (void);
static void advertising_init (void);
static void services_init    (void);
static void conn_params_init (void);
static void sec_params_init  (void);
                                                        /* Event dispatchers */
static void ble_evt_dispatch (ble_evt_t *p_ble_evt);
static void sys_evt_dispatch (uint32_t   sys_evt);
                                                           /* Event handlers */
static void on_adv_evt         (ble_adv_evt_t ble_adv_evt);
static void on_conn_params_evt (ble_conn_params_evt_t *p_evt);
static void on_ble_evt         (ble_evt_t *p_ble_evt);
static void on_hts_evt         (ble_hts_t *p_hts, ble_hts_evt_t *p_evt);
static void on_mds_evt         (ble_mds_t *p_mds, uint8_t *data, uint16_t len);
                                                           /* Error handlers */
static void assert_nrf_callback (uint16_t line_num, const uint8_t *p_file_name);
static void service_error_handler     (uint32_t nrf_error);
static void conn_params_error_handler (uint32_t nrf_error);
                                                           /* Little helpers */
static void sleep_mode_enter             (void);
// static void temperature_measurement_send (void);
static void showConnect                  (bool connect);
                                                        /* Communicaton task */
static void StackTask (void);

/*---- Implementation -------------------------------------------------------*/

/*.... Public functions .....................................................*/

void STACK_Init (uint8_t Prio)
{
    APP_TIMER_INIT (TIMER_PRESC, MAX_TIMERS, TIMER_QUEUE_SIZE, false);

    OS_CREATETASK (&mTaskCb, "CommsTask", StackTask, Prio, Stack);
}


/*.... Private functions ....................................................*/
/*     Initialisation routines                                               */

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init (void)
{
    uint32_t             err_code;
    uint32_t             appRamBase = 0x20001E10;                                             // !!! HACK !!! value was given back by sd_ble_enable
    ble_enable_params_t  ep;


    memset (&ep, 0, sizeof (ep));

  #ifdef S130
    ep.gatts_enable_params.attr_tab_size = BLE_GATTS_ATTR_TAB_SIZE_DEFAULT;
  #endif

    ep.gap_enable_params.periph_conn_count = 1;
    ep.gatts_enable_params.service_changed = IS_SRVC_CHANGED_CHARACT_PRESENT;
    err_code = sd_ble_enable (&ep, &appRamBase);
    APP_ERROR_CHECK (err_code);

    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_ble_evt_handler_set (ble_evt_dispatch);
    APP_ERROR_CHECK (err_code);

    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_sys_evt_handler_set (sys_evt_dispatch);
    APP_ERROR_CHECK (err_code);

  // !!! HACK !!! overwrite the interrupt priority
    NVIC_SetPriority (SWI2_EGU2_IRQn, APP_IRQ_PRIORITY_LOWEST);
}

/**@brief Function for the GAP initialization.
 *
 * @details This function sets up all the necessary GAP (Generic Access Profile) parameters of the
 *          device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init (void)
{
    uint32_t                 err_code;
    ble_gap_conn_params_t    gap_conn_params;
    ble_gap_conn_sec_mode_t  sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN (&sec_mode);

    err_code = sd_ble_gap_device_name_set (&sec_mode, (const uint8_t *)DEVICE_NAME, strlen (DEVICE_NAME));
    APP_ERROR_CHECK (err_code);

    err_code = sd_ble_gap_appearance_set (BLE_APPEARANCE_UNKNOWN);
    APP_ERROR_CHECK (err_code);

    memset (&gap_conn_params, 0, sizeof (gap_conn_params));

    gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency     = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set (&gap_conn_params);
    APP_ERROR_CHECK (err_code);
}

/**@brief Function for initializing the Advertising functionality.
 */
static void advertising_init (void)
{
    uint32_t       err_code;
    ble_advdata_t  advdata;
    ble_advdata_t  scanrsp;

    // Build advertising data struct to pass into @ref ble_advertising_init.
    memset (&advdata, 0, sizeof (advdata));
    advdata.name_type          = BLE_ADVDATA_FULL_NAME;
    advdata.include_appearance = false;
    advdata.flags              = BLE_GAP_ADV_FLAGS_LE_ONLY_LIMITED_DISC_MODE;

    memset (&scanrsp, 0, sizeof (scanrsp));
    scanrsp.uuids_complete.uuid_cnt = sizeof (m_adv_uuids) / sizeof (m_adv_uuids[0]);
    scanrsp.uuids_complete.p_uuids  = m_adv_uuids;

    ble_adv_modes_config_t options = {0};
    options.ble_adv_fast_enabled  = BLE_ADV_FAST_ENABLED;
    options.ble_adv_fast_interval = APP_ADV_INTERVAL;
    options.ble_adv_fast_timeout  = APP_ADV_TIMEOUT_IN_SECONDS;

    err_code = ble_advertising_init (&advdata, &scanrsp, &options, on_adv_evt, NULL);
    APP_ERROR_CHECK (err_code);
}

/**@brief Function for initializing services that will be used by the application.
 */
static void services_init (void)
{
    uint32_t         err_code;
    ble_hts_init_t   hts_init;
    ble_bas_init_t   bas_init;
    ble_dis_init_t   dis_init;
    ble_dis_sys_id_t sys_id;
    ble_mds_init_t   mds_init;

    // Initialise Health Thermometer Service
    memset (&hts_init, 0, sizeof (hts_init));

    hts_init.evt_handler                 = on_hts_evt;
    hts_init.temp_type_as_characteristic = TEMP_TYPE_AS_CHARACTERISTIC;
    hts_init.temp_type                   = BLE_HTS_TEMP_TYPE_BODY;

    // Here the sec level for the Health Thermometer Service can be changed/increased.
    BLE_GAP_CONN_SEC_MODE_SET_ENC_NO_MITM (&hts_init.hts_meas_attr_md.cccd_write_perm);
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS   (&hts_init.hts_meas_attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS   (&hts_init.hts_meas_attr_md.write_perm);

    BLE_GAP_CONN_SEC_MODE_SET_OPEN        (&hts_init.hts_temp_type_attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS   (&hts_init.hts_temp_type_attr_md.write_perm);

    err_code = ble_hts_init (&m_hts, &hts_init);
    APP_ERROR_CHECK (err_code);

    // Initialise Battery Service
    memset (&bas_init, 0, sizeof (bas_init));

    // Here the sec level for the Battery Service can be changed/increased.
    BLE_GAP_CONN_SEC_MODE_SET_OPEN      (&bas_init.battery_level_char_attr_md.cccd_write_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN      (&bas_init.battery_level_char_attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS (&bas_init.battery_level_char_attr_md.write_perm);

    BLE_GAP_CONN_SEC_MODE_SET_OPEN (&bas_init.battery_level_report_read_perm);

    bas_init.evt_handler          = NULL;
    bas_init.support_notification = true;
    bas_init.p_report_ref         = NULL;
    bas_init.initial_batt_level   = 100;

    err_code = ble_bas_init (&m_bas, &bas_init);
    APP_ERROR_CHECK (err_code);

    // Initialise Device Information Service
    memset (&dis_init, 0, sizeof (dis_init));

    ble_srv_ascii_to_utf8 (&dis_init.manufact_name_str, MANUFACTURER_NAME);
    ble_srv_ascii_to_utf8 (&dis_init.model_num_str,     MODEL_NUM);

    sys_id.manufacturer_id            = MANUFACTURER_ID;
    sys_id.organizationally_unique_id = ORG_UNIQUE_ID;
    dis_init.p_sys_id                 = &sys_id;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN      (&dis_init.dis_attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS (&dis_init.dis_attr_md.write_perm);

    err_code = ble_dis_init (&dis_init);
    APP_ERROR_CHECK (err_code);

    // Initialise Device Service
    memset (&mds_init, 0, sizeof (mds_init));
    mds_init.data_handler = on_mds_evt;

    err_code = ble_mds_init (&m_mds, &mds_init);
    APP_ERROR_CHECK (err_code);
}

/**@brief Function for initializing the Connection Parameters module.
 */
static void conn_params_init (void)
{
    uint32_t                err_code;
    ble_conn_params_init_t  cp_init;

    memset (&cp_init, 0, sizeof(cp_init));

    cp_init.p_conn_params                  = NULL;
    cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
    cp_init.next_conn_params_update_delay  = NEXT_CONN_PARAMS_UPDATE_DELAY;
    cp_init.max_conn_params_update_count   = MAX_CONN_PARAMS_UPDATE_COUNT;
    cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
    cp_init.disconnect_on_fail             = false;
    cp_init.evt_handler                    = on_conn_params_evt;
    cp_init.error_handler                  = conn_params_error_handler;

    err_code = ble_conn_params_init (&cp_init);
    APP_ERROR_CHECK (err_code);
}

/**@brief Function for initializing security parameters.
 */
static void sec_params_init (void)
{
    m_sec_params.bond         = SEC_PARAM_BOND;
    m_sec_params.mitm         = SEC_PARAM_MITM;
    m_sec_params.io_caps      = SEC_PARAM_IO_CAPABILITIES;
    m_sec_params.oob          = SEC_PARAM_OOB;
    m_sec_params.min_key_size = SEC_PARAM_MIN_KEY_SIZE;
    m_sec_params.max_key_size = SEC_PARAM_MAX_KEY_SIZE;
}


/*     Event dispatchers                                                     */

/**@brief Function for dispatching a BLE stack event to all modules with a BLE stack event handler.
 *
 * @details This function is called from the scheduler in the main loop after a BLE stack
 *          event has been received.
 *
 * @param[in] p_ble_evt  Bluetooth stack event.
 */
static void ble_evt_dispatch (ble_evt_t *p_ble_evt)
{
    ble_hts_on_ble_evt (&m_hts, p_ble_evt);
    ble_bas_on_ble_evt (&m_bas, p_ble_evt);
    ble_mds_on_ble_evt (&m_mds, p_ble_evt);

    on_ble_evt (p_ble_evt);

    ble_conn_params_on_ble_evt (p_ble_evt);
    ble_advertising_on_ble_evt (p_ble_evt);
}

/**@brief Function for dispatching a system event to interested modules.
 *
 * @details This function is called from the System event interrupt handler after a system
 *          event has been received.
 *
 * @param[in] sys_evt  System stack event.
 */
static void sys_evt_dispatch (uint32_t sys_evt)
{
    ble_advertising_on_sys_evt (sys_evt);
}


/*     Event handlers                                                        */

/**@brief Function for handling advertising events.
 *
 * @details This function will be called for advertising events which are passed to the application.
 *
 * @param[in] ble_adv_evt  Advertising event.
 */
static void on_adv_evt (ble_adv_evt_t ble_adv_evt)
{
    uint32_t err_code;

    switch (ble_adv_evt)
    {
        case BLE_ADV_EVT_FAST:
//            err_code = bsp_indication_set(BSP_INDICATE_ADVERTISING);
//            APP_ERROR_CHECK(err_code);
            break;

        case BLE_ADV_EVT_IDLE:
//            sleep_mode_enter ();
            break;

        default:
            break;
    }
}

/**@brief Function for handling the Connection Parameters Module.
 *
 * @details This function will be called for all events in the Connection Parameters Module which
 *          are passed to the application.
 *          @note All this function does is to disconnect. This could have been done by simply
 *                setting the disconnect_on_fail config parameter, but instead we use the event
 *                handler mechanism to demonstrate its use.
 *
 * @param[in] p_evt  Event received from the Connection Parameters Module.
 */
static void on_conn_params_evt (ble_conn_params_evt_t *p_evt)
{
    uint32_t  err_code;

    if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
    {
        err_code = sd_ble_gap_disconnect (m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
        APP_ERROR_CHECK (err_code);
    }
}

/**@brief Function for handling the Application's BLE Stack events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void on_ble_evt (ble_evt_t *p_ble_evt)
{
    uint32_t                         err_code;
    static ble_gap_evt_auth_status_t m_auth_status;
    bool                             master_id_matches;
    ble_gap_sec_kdist_t             *p_distributed_keys;
    ble_gap_enc_info_t              *p_enc_info;
    ble_gap_irk_t                   *p_id_info;
    ble_gap_sign_info_t             *p_sign_info;

    static ble_gap_enc_key_t         m_enc_key;           /**< Encryption Key (Encryption Info and Master ID). */
    static ble_gap_id_key_t          m_id_key;            /**< Identity Key (IRK and address). */
    static ble_gap_sign_info_t       m_sign_key;          /**< Signing Key (Connection Signature Resolving Key). */
    static ble_gap_sec_keyset_t      m_keys = {.keys_periph = {&m_enc_key, &m_id_key, &m_sign_key}};

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            showConnect (true);
            m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            showConnect (false);
            m_conn_handle = BLE_CONN_HANDLE_INVALID;
            break;

        case BLE_GATTS_EVT_TIMEOUT:
            if (p_ble_evt->evt.gatts_evt.params.timeout.src == BLE_GATT_TIMEOUT_SRC_PROTOCOL)
            {
                err_code = sd_ble_gap_disconnect (m_conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
                APP_ERROR_CHECK(err_code);
            }
            break;

        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
            err_code = sd_ble_gap_sec_params_reply (m_conn_handle, BLE_GAP_SEC_STATUS_SUCCESS, &m_sec_params, &m_keys);
            APP_ERROR_CHECK (err_code);
            break;

//        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
//            err_code = sd_ble_gatts_sys_attr_set (m_conn_handle, NULL, 0, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS | BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
//            APP_ERROR_CHECK (err_code);
//            break;

        case BLE_GAP_EVT_AUTH_STATUS:
            m_auth_status = p_ble_evt->evt.gap_evt.params.auth_status;
            break;

        case BLE_GAP_EVT_SEC_INFO_REQUEST:
            master_id_matches  = memcmp (&p_ble_evt->evt.gap_evt.params.sec_info_request.master_id, &m_enc_key.master_id, sizeof (ble_gap_master_id_t)) == 0;
            p_distributed_keys = &m_auth_status.kdist_periph;

            p_enc_info  = (p_distributed_keys->enc  && master_id_matches) ? &m_enc_key.enc_info : NULL;
            p_id_info   = (p_distributed_keys->id   && master_id_matches) ? &m_id_key.id_info   : NULL;
            p_sign_info = (p_distributed_keys->sign && master_id_matches) ? &m_sign_key         : NULL;

            err_code = sd_ble_gap_sec_info_reply (m_conn_handle, p_enc_info, p_id_info, p_sign_info);
            APP_ERROR_CHECK(err_code);
            break;

        default:
            // No implementation needed.
            break;
    }
}

/**@brief Function for handling the Health Thermometer Service events.
 *
 * @details This function will be called for all Health Thermometer Service events which are passed
 *          to the application.
 *
 * @param[in] p_hts  Health Thermometer Service structure.
 * @param[in] p_evt  Event received from the Health Thermometer Service.
 */
static void on_hts_evt (ble_hts_t *p_hts, ble_hts_evt_t *p_evt)
{
    switch (p_evt->evt_type)
    {
        case BLE_HTS_EVT_INDICATION_ENABLED:
            // Indication has been enabled, send a single temperature measurement
            temperature_measurement_send ();
            break;

        case BLE_HTS_EVT_INDICATION_CONFIRMED:
            m_hts_meas_ind_conf_pending = false;
            break;

        default:
            // No implementation needed.
            break;
    }
}

/**@brief Function for handling the Device Communication (MDS) events.
 *
 * @param[in]   p_mds   MDS Event
 * @param[in]   data    data received
 * @param[in]   len     length of data received
 */
static void on_mds_evt (ble_mds_t *p_mds, uint8_t *data, uint16_t len)
{

}


/*     Error handlers                                                        */

/**@brief Callback function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num   Line number of the failing ASSERT call.
 * @param[in] file_name  File name of the failing ASSERT call.
 */
static void assert_nrf_callback (uint16_t line_num, const uint8_t *p_file_name)
{
    app_error_handler (DEAD_BEEF, line_num, p_file_name);
}

/**@brief Function for handling Service errors.
 *
 * @details A pointer to this function will be passed to each service which may need to inform the
 *          application about an error.
 *
 * @param[in] nrf_error  Error code containing information about what went wrong.
 */
static void service_error_handler (uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}

/**@brief Function for handling a Connection Parameters error.
 *
 * @param[in] nrf_error  Error code containing information about what went wrong.
 */
static void conn_params_error_handler (uint32_t nrf_error)
{
    APP_ERROR_HANDLER (nrf_error);
}


/*     Little helpers                                                        */

/**@brief Function for putting the chip into sleep mode.
 *
 * @note This function will not return.
 */
static void sleep_mode_enter (void)
{
    unsigned long  err_code;

    err_code = sd_power_system_off ();
    APP_ERROR_CHECK (err_code);
}

/**@brief Function for simulating and sending one Temperature Measurement.
 */
void temperature_measurement_send (void)
{
    uint32_t        err_code;
    ble_hts_meas_t  sim;
    ble_date_time_t ts = { 2012, 12, 5, 11, 50, 0 };

    if (m_hts_meas_ind_conf_pending)
        return;

    sim.temp_in_fahr_units = false;                                        /**< True if Temperature is in Fahrenheit units, Celcius otherwise. */
    sim.time_stamp_present = false;                                        /**< True if Time Stamp is present. */
    sim.temp_type_present = (TEMP_TYPE_AS_CHARACTERISTIC ? false : true);  /**< True if Temperature Type is present. */
    sim.temp_in_celcius.exponent = -2;
    sim.temp_in_celcius.mantissa = celsiusX100;
    sim.temp_in_fahr.exponent    = -2;
    sim.temp_in_fahr.mantissa    = (32 * 100) + ((celsiusX100 * 9) / 5);
    sim.time_stamp = ts;                                                   /**< Time Stamp. */
    sim.temp_type = BLE_HTS_TEMP_TYPE_BODY;                                /**< Temperature Type. */

    err_code = ble_hts_measurement_send (&m_hts, &sim);

    switch (err_code)
    {
        case NRF_SUCCESS:
        // Measurement was successfully sent, wait for confirmation.
        m_hts_meas_ind_conf_pending = true;
        break;

        case NRF_ERROR_INVALID_STATE:
        // Ignore error.
        break;

        default:
        // APP_ERROR_HANDLER(err_code);
        break;
    }

    celsiusX100 += 50;
}


void mds_send (char *data, unsigned char len)
{
    ble_mds_response_send (&m_mds, data, len);
}


static void  showConnect (bool connect)
{
    if (connect)  nrf_gpio_pin_set   (LED_1_PIN);
        else      nrf_gpio_pin_clear (LED_1_PIN);
}


/*     Communication task                                                    */

static void StackTask (void)
{
    bool      ind      = false;
    bool      tickTock = true;
    uint32_t  err_code;

    // Initialize
    ble_stack_init   ();
    gap_params_init  ();
    services_init    ();
    advertising_init ();
    conn_params_init ();
    sec_params_init  ();

    // Start execution
    err_code = ble_advertising_start (BLE_ADV_MODE_FAST);
    APP_ERROR_CHECK (err_code);

    showConnect (false);
    ble_bas_battery_level_update  (&m_bas, 63);
    ble_hts_is_indication_enabled (&m_hts, &ind);

    while (true)                                             // Enter main loop
    {
        err_code = sd_app_evt_wait ();
        APP_ERROR_CHECK (err_code);

        if (tickTock)  nrf_gpio_pin_set   (LED_3_PIN);
            else       nrf_gpio_pin_clear (LED_3_PIN);

        tickTock = ! tickTock;
    }
}
